"""Package constants."""
# pylint: disable=too-few-public-methods

from enum import IntEnum


class ColorSpace(IntEnum):
    """Color space specifier."""

    SRGB = 1
    "sRBG"

    UNCALIBRATED = 0xFFFF
    "Uncalibrated or Other"


class ExifMarkers:
    """EXIF marker segments bytes."""

    SEG_PREFIX = b"\xff"
    """Generic Segment Prefix"""

    SOI = SEG_PREFIX + b"\xd8"
    """Start of Image"""

    APP1 = SEG_PREFIX + b"\xe1"
    """EXIF Attribute Information (Application Segment 1)"""

    APP2 = SEG_PREFIX + b"\xe2"
    """EXIF Extended Data (Application Segment 2)"""

    DQT = SEG_PREFIX + b"\xdb"
    """Quantization Table Definition"""

    DHT = SEG_PREFIX + b"\xc4"
    """Huffman Table Definition"""

    DRI = SEG_PREFIX + b"\xdd"
    """Restart Interoperability Definition"""

    SOF = SEG_PREFIX + b"\xc0"
    """Start of Frame"""

    SOS = SEG_PREFIX + b"\xda"
    """Start of Scan"""

    EOI = SEG_PREFIX + b"\xd9"
    """End of Image"""


class ExifTypes(IntEnum):
    """EXIF datatype indicator for IFD structure."""

    BYTE = 1
    """8-Bit Unsigned Integer"""

    ASCII = 2
    """Null-Terminated ASCII Codes"""

    SHORT = 3
    """16-Bit Unsigned Integer"""

    LONG = 4
    """32-Bit Unsigned Integer"""

    RATIONAL = 5
    """Two (Numerator and Denominator) LONGs"""

    SSHORT = 8
    """16-Bit Signed Integer"""

    SLONG = 9
    """32-Bit Signed Integer"""

    SRATIONAL = 10
    """Two (Numerator and Denominator) SLONGs"""


class ExposureMode(IntEnum):
    """Exposure mode set when the image was shot."""

    AUTO_EXPOSURE = 0
    """Auto Exposure"""

    MANUAL_EXPOSURE = 1
    """Manual Exposure"""

    AUTO_BRACKET = 2
    """Auto Bracket"""


class ExposureProgram(IntEnum):
    """Class of the program used by the camera to set exposure when the picture is taken."""

    NOT_DEFINED = 0
    """Not Defined"""

    MANUAL = 1
    """Manual"""

    NORMAL_PROGRAM = 2
    """Normal Program"""

    APERTURE_PRIORITY = 3
    """Aperture Priority"""

    SHUTTER_PRIORITY = 4
    """Shutter Priority"""

    CREATIVE_PROGRAM = 5
    """Creative Program (Biased Toward Depth of Field)"""

    ACTION_PROGRAM = 6
    """Action Program (Biased Toward Fast Shutter Speed)"""

    PORTRAIT_MODE = 7
    """Portrait Mode (For Closeup Photos with the Background out of Focus)"""

    LANDSCAPE_MODE = 8
    """Landscape Kode (For Landscape Photos with the Background in Focus)"""


class GpsAltitudeRef(IntEnum):
    """Altitude used as the reference altitude."""

    ABOVE_SEA_LEVEL = 0
    """Above Sea Level"""

    BELOW_SEA_LEVEL = 1
    """Below Sea Level"""


class LightSource(IntEnum):
    """Class of the program used by the camera to set exposure when the picture is taken."""

    UNKNOWN = 0
    """Unknown"""

    DAYLIGHT = 1
    """Daylight"""

    FLUORESCENT = 2
    """Fluorescent"""

    TUNGSTEN = 3
    """Tungsten (Incandescent Light)"""

    FLASH = 4
    """Flash"""

    FINE_WEATHER = 9
    """Fine Weather"""

    CLOUDY_WEATHER = 10
    """Cloudy Weather"""

    SHADE = 11
    """Shade"""

    DAYLIGHT_FLUORESCENT = 12
    """Daylight Fluorescent (D 5700 - 7100K)"""

    DAY_WHITE_FLUORESCENT = 13
    """Day White Fluorescent (N 4600 - 5400K)"""

    COOL_WHITE_FLUORESCENT = 14
    """Cool White Fluorescent (W 3900 - 4500K)"""

    WHITE_FLUORESCENT = 15
    """White Fluorescent (WW 3200 - 3700K)"""

    STANDARD_LIGHT_A = 17
    """Standard Light A"""

    STANDARD_LIGHT_B = 18
    """Standard Light B"""

    STANDARD_LIGHT_C = 19
    """Standard Light C"""

    D55 = 20
    """D55"""

    D65 = 21
    """D65"""

    D75 = 22
    """D75"""

    D50 = 23
    """D50"""

    ISO_STUDIO_TUNGSTEN = 24
    """ISO Studio Tungsten"""

    OTHER = 255
    """Other Light Source"""


class MeteringMode(IntEnum):
    """Metering mode."""

    UNKNOWN = 0
    """Unknown"""

    AVERAGE = 1
    """Average"""

    CENTER_WEIGHTED_AVERAGE = 2
    """Center Weighted Average"""

    SPOT = 3
    """Spot"""

    MULTI_SPOT = 4
    """Multi Spot"""

    PATTERN = 5
    """Pattern"""

    PARTIAL = 6
    """Partial"""

    OTHER = 255
    """Other"""


class Orientation(IntEnum):
    """Image orientation in terms of rows and columns."""

    TOP_LEFT = 1
    """The 0th row is at the visual top of the image and the 0th column is the visual
    left-hand side."""

    TOP_RIGHT = 2
    """The 0th row is at the visual top of the image and the 0th column is the visual
    right-hand side."""

    BOTTOM_RIGHT = 3
    """The 0th row is at the visual bottom of the image and the 0th column is the visual
    right-hand side."""

    BOTTOM_LEFT = 4
    """The 0th row is at the visual bottom of the image and the 0th column is the visual
    left-hand side."""

    LEFT_TOP = 5
    """The 0th row is the visual left-hand side of the image and the 0th column is the
    visual top."""

    RIGHT_TOP = 6
    """The 0th row is the visual right-hand side of the image and the 0th column is the
    visual bottom."""

    RIGHT_BOTTOM = 7
    """The 0th row is the visual right-hand side of the image and the 0th column is the
    visual bottom."""

    LEFT_BOTTOM = 8
    """The 0th row is the visual left-hand side of the image and the 0th column is the
    visual bottom."""


class ResolutionUnit(IntEnum):
    """Unit for measuring X resolution and Y resolution tags."""

    INCHES = 2
    """Inches or Unknown"""

    CENTIMETERS = 3
    """Centimeters"""


class Saturation(IntEnum):
    """Saturation processing applied by camera."""

    NORMAL = 0
    """Normal Saturation"""

    LOW = 1
    """Low Saturation"""

    HIGH = 2
    """High Saturation"""


class SceneCaptureType(IntEnum):
    """Type of scene that was shot or the mode in which the image was shot."""

    STANDARD = 0
    """Standard"""

    LANDSCAPE = 1
    """Landscape"""

    PORTRAIT = 2
    """Portrait"""

    NIGHT_SCENE = 3
    """Night Scene"""


class SensingMethod(IntEnum):
    """Image sensor type on the camera or input device."""

    NOT_DEFINED = 1
    """Not Defined"""

    ONE_CHIP_COLOR_AREA_SENSOR = 2
    """One-Chip Color Area Sensor"""

    TWO_CHIP_COLOR_AREA_SENSOR = 3
    """Two-Chip Color Area Sensor"""

    THREE_CHIP_COLOR_AREA_SENSOR = 4
    """Three-Chip Color Area Sensor"""

    COLOR_SEQUENTIAL_AREA_SENSOR = 5
    """Color Sequential Area Sensor"""

    TRILINEAR_SENSOR = 7
    """Trilinear Sensor"""

    COLOR_SEQUENTIAL_LINEAR_SENSOR = 8
    """Color Sequential Linear Sensor"""


class Sharpness(IntEnum):
    """Sharpness processing applied by camera."""

    NORMAL = 0
    """Normal"""

    SOFT = 1
    """Soft"""

    HARD = 2
    """Hard"""


class WhiteBalance(IntEnum):
    """White balance mode set when the image was shot."""

    AUTO = 0
    """Auto White Balance"""

    MANUAL = 1
    """Manual White Balance"""


ATTRIBUTE_ID_MAP = {
    # TIFF Rev. 6.0 Section A: Image Data Structure Tags
    "image_width": 256,
    "image_height": 257,
    "bits_per_sample": 258,
    "compression": 259,
    "photometric_interpretation": 262,
    "orientation": 274,
    "samples_per_pixel": 277,
    "planar_configuration": 284,
    "subsampling_ratio_of_y_to_c": 530,
    "y_and_c_positioning": 531,
    "x_resolution": 282,
    "y_resolution": 283,
    "resolution_unit": 296,
    # TIFF Rev. 6.0 Section B: Recording Offset Tags
    "strip_offsets": 273,
    "rows_per_strip": 278,
    "strip_byte_counts": 279,
    "jpeg_interchange_format": 513,
    "jpeg_interchange_format_length": 514,
    # TIFF Rev. 6.0 Section C: Image Data Characteristic Tags
    "transfer_function": 301,
    "white_point": 318,
    "primary_chromaticities": 319,
    "matrix_coefficients": 529,
    "reference_black_white": 532,
    # TIFF Rev. 6.0 Section D: Other Tags
    "datetime": 306,
    "image_description": 270,
    "make": 271,
    "model": 272,
    "software": 305,
    "artist": 315,
    "copyright": 33432,
    # EXIF Tags
    "exposure_time": 33434,
    "f_number": 33437,
    "exposure_program": 34850,
    "spectral_sensitivity": 34852,
    "photographic_sensitivity": 34855,
    "oecf": 34856,
    "sensitivity_type": 34864,
    "standard_output_sensitivity": 34865,
    "recommended_exposure_index": 34866,
    "iso_speed": 34867,
    "iso_speed_latitude_yyy": 34868,
    "iso_speed_latitude_zzz": 34869,
    "exif_version": 36864,
    "datetime_original": 36867,
    "datetime_digitized": 36868,
    "offset_time": 36880,
    "offset_time_original": 36881,
    "offset_time_digitized": 36882,
    "components_configuration": 37121,
    "compressed_bits_per_pixel": 37122,
    "shutter_speed_value": 37377,
    "aperture_value": 37378,
    "brightness_value": 37379,
    "exposure_bias_value": 37380,
    "max_aperture_value": 37381,
    "subject_distance": 37382,
    "metering_mode": 37383,
    "light_source": 37384,
    "flash": 37385,
    "focal_length": 37386,
    "subject_area": 37396,
    "maker_note": 37500,
    "user_comment": 37510,
    "subsec_time": 37520,
    "subsec_time_original": 37521,
    "subsec_time_digitized": 37522,
    "temperature": 37888,
    "humidity": 37889,
    "pressure": 37890,
    "water_depth": 37891,
    "acceleration": 37892,
    "camera_elevation_angle": 37893,
    "xp_title": 0x9C9B,
    "xp_comment": 0x9C9C,
    "xp_author": 0x9C9D,
    "xp_keywords": 0x9C9E,
    "xp_subject": 0x9C9F,
    "flashpix_version": 40960,
    "color_space": 40961,
    "pixel_x_dimension": 40962,
    "pixel_y_dimension": 40963,
    "related_sound_file": 40964,
    "flash_energy": 41483,
    "spatial_frequency_response": 41484,
    "focal_plane_x_resolution": 41486,
    "focal_plane_y_resolution": 41487,
    "focal_plane_resolution_unit": 41488,
    "subject_location": 41492,
    "exposure_index": 41493,
    "sensing_method": 41495,
    "file_source": 41728,
    "scene_type": 41729,
    "cfa_pattern": 41730,
    "custom_rendered": 41985,
    "exposure_mode": 41986,
    "white_balance": 41987,
    "digital_zoom_ratio": 41988,
    "focal_length_in_35mm_film": 41989,
    "scene_capture_type": 41990,
    "gain_control": 41991,
    "contrast": 41992,
    "saturation": 41993,
    "sharpness": 41994,
    "device_setting_description": 41995,
    "subject_distance_range": 41996,
    "image_unique_id": 42016,
    "camera_owner_name": 42032,
    "body_serial_number": 42033,
    "lens_specification": 42034,
    "lens_make": 42035,
    "lens_model": 42036,
    "lens_serial_number": 42037,
    "gamma": 42240,
    # GPS Info Tags
    "gps_version_id": 0,
    "gps_latitude_ref": 1,
    "gps_latitude": 2,
    "gps_longitude_ref": 3,
    "gps_longitude": 4,
    "gps_altitude_ref": 5,
    "gps_altitude": 6,
    "gps_timestamp": 7,
    "gps_satellites": 8,
    "gps_status": 9,
    "gps_measure_mode": 10,
    "gps_dop": 11,
    "gps_speed_ref": 12,
    "gps_speed": 13,
    "gps_track_ref": 14,
    "gps_track": 15,
    "gps_img_direction_ref": 16,
    "gps_img_direction": 17,
    "gps_map_datum": 18,
    "gps_dest_latitude_ref": 19,
    "gps_dest_latitude": 20,
    "gps_dest_longitude_ref": 21,
    "gps_dest_longitude": 22,
    "gps_dest_bearing_ref": 23,
    "gps_dest_bearing": 24,
    "gps_dest_distance_ref": 25,
    "gps_dest_distance": 26,
    "gps_processing_method": 27,
    "gps_area_information": 28,
    "gps_datestamp": 29,
    "gps_differential": 30,
    "gps_horizontal_positioning_error": 31,
    # EXIF-specific IFDs (for __dir__ Display)
    "_exif_ifd_pointer": 34665,
    "_gps_ifd_pointer": 34853,
    "_interoperability_ifd_Pointer": 40965,
}

ATTRIBUTE_NAME_MAP = {value: key for key, value in ATTRIBUTE_ID_MAP.items()}

ATTRIBUTE_TYPE_MAP = (
    {  # tuple of type ID and IFD number description used when adding new tags
        "aperture_value": (int(ExifTypes.RATIONAL), "exif"),
        "artist": (int(ExifTypes.ASCII), 0),
        "brightness_value": (int(ExifTypes.SRATIONAL), "exif"),
        "color_space": (int(ExifTypes.SHORT), "exif"),
        "contrast": (int(ExifTypes.SHORT), "exif"),
        "copyright": (int(ExifTypes.ASCII), 0),
        "custom_rendered": (int(ExifTypes.SHORT), "exif"),
        "datetime": (int(ExifTypes.ASCII), 0),
        "datetime_digitized": (int(ExifTypes.ASCII), "exif"),
        "datetime_original": (int(ExifTypes.ASCII), "exif"),
        "digital_zoom_ratio": (int(ExifTypes.RATIONAL), "exif"),
        "exposure_bias_value": (int(ExifTypes.SRATIONAL), "exif"),
        "exposure_index": (int(ExifTypes.RATIONAL), "exif"),
        "exposure_mode": (int(ExifTypes.SHORT), "exif"),
        "exposure_program": (int(ExifTypes.SHORT), "exif"),
        "exposure_time": (int(ExifTypes.RATIONAL), "exif"),
        "f_number": (int(ExifTypes.RATIONAL), "exif"),
        "flash": (int(ExifTypes.SHORT), "exif"),
        "flash_energy": (int(ExifTypes.RATIONAL), "exif"),
        "focal_length": (int(ExifTypes.RATIONAL), "exif"),
        "focal_length_in_35mm_film": (int(ExifTypes.SHORT), "exif"),
        "focal_plane_resolution_unit": (int(ExifTypes.SHORT), "exif"),
        "gain_control": (int(ExifTypes.RATIONAL), "exif"),
        "gps_altitude": (int(ExifTypes.RATIONAL), "gps"),
        "gps_altitude_ref": (int(ExifTypes.BYTE), "gps"),
        "gps_datestamp": (int(ExifTypes.ASCII), "gps"),
        "gps_dest_bearing": (int(ExifTypes.RATIONAL), "gps"),
        "gps_dest_bearing_ref": (int(ExifTypes.ASCII), "gps"),
        "gps_dest_distance": (int(ExifTypes.RATIONAL), "gps"),
        "gps_dest_distance_ref": (int(ExifTypes.ASCII), "gps"),
        "gps_dest_latitude": (int(ExifTypes.RATIONAL), "gps"),
        "gps_dest_latitude_ref": (int(ExifTypes.ASCII), "gps"),
        "gps_dest_longitude": (int(ExifTypes.RATIONAL), "gps"),
        "gps_dest_longitude_ref": (int(ExifTypes.ASCII), "gps"),
        "gps_differential": (int(ExifTypes.SHORT), "gps"),
        "gps_dop": (int(ExifTypes.RATIONAL), "gps"),
        "gps_img_direction": (int(ExifTypes.RATIONAL), "gps"),
        "gps_img_direction_ref": (int(ExifTypes.ASCII), "gps"),
        "gps_latitude": (int(ExifTypes.RATIONAL), "gps"),
        "gps_latitude_ref": (int(ExifTypes.ASCII), "gps"),
        "gps_longitude": (int(ExifTypes.RATIONAL), "gps"),
        "gps_longitude_ref": (int(ExifTypes.ASCII), "gps"),
        "gps_map_datum": (int(ExifTypes.ASCII), "gps"),
        "gps_measure_mode": (int(ExifTypes.ASCII), "gps"),
        "gps_satellites": (int(ExifTypes.ASCII), "gps"),
        "gps_speed": (int(ExifTypes.RATIONAL), "gps"),
        "gps_speed_ref": (int(ExifTypes.ASCII), "gps"),
        "gps_status": (int(ExifTypes.ASCII), "gps"),
        "gps_timestamp": (int(ExifTypes.RATIONAL), "gps"),
        "gps_track": (int(ExifTypes.RATIONAL), "gps"),
        "gps_track_ref": (int(ExifTypes.ASCII), "gps"),
        "gps_version_id": (int(ExifTypes.BYTE), "gps"),
        "image_description": (int(ExifTypes.ASCII), 0),
        "image_unique_id": (int(ExifTypes.ASCII), "exif"),
        "light_source": (int(ExifTypes.SHORT), "exif"),
        "make": (int(ExifTypes.ASCII), 0),
        "max_aperture_value": (int(ExifTypes.RATIONAL), 0),
        "metering_mode": (int(ExifTypes.SHORT), "exif"),
        "model": (int(ExifTypes.ASCII), 0),
        "orientation": (int(ExifTypes.SHORT), 0),
        "pixel_x_dimension": (int(ExifTypes.SHORT), "exif"),
        "pixel_y_dimension": (int(ExifTypes.SHORT), "exif"),
        "saturation": (int(ExifTypes.SHORT), "exif"),
        "scene_capture_type": (int(ExifTypes.SHORT), "exif"),
        "sensing_method": (int(ExifTypes.SHORT), "exif"),
        "shutter_speed_value": (int(ExifTypes.SRATIONAL), "exif"),
        "software": (int(ExifTypes.ASCII), 0),
        "sharpness": (int(ExifTypes.SHORT), "exif"),
        "spectral_sensitivity": (int(ExifTypes.ASCII), "exif"),
        "subsec_time": (int(ExifTypes.ASCII), "exif"),
        "subsec_time_original": (int(ExifTypes.ASCII), "exif"),
        "subsec_time_digitized": (int(ExifTypes.ASCII), "exif"),
        "subject_distance": (int(ExifTypes.RATIONAL), "exif"),
        "subject_distance_range": (int(ExifTypes.SHORT), "exif"),
        "subject_location": (int(ExifTypes.SHORT), "exif"),
        "user_comment": (
            7,
            "exif",
        ),  # 7 is UNDEFINED (if packed in enum, it would cause downstream issues)
        "white_balance": (int(ExifTypes.SHORT), "exif"),
        "_exif_ifd_pointer": (int(ExifTypes.LONG), 0),
        "_gps_ifd_pointer": (int(ExifTypes.LONG), 0),
    }
)


ERROR_IMG_NO_ATTR = "image does not have attribute {0}"



"""APP1 metadata interface module for EXIF tags."""

import warnings
from typing import List

from plum.buffer import Buffer
from plum.bigendian import uint16
from plum.exceptions import UnpackError

from exif._utils import pack_into, value_fits_in_ifd_tag
from exif._constants import (
    ATTRIBUTE_ID_MAP,
    ATTRIBUTE_NAME_MAP,
    ATTRIBUTE_TYPE_MAP,
    ERROR_IMG_NO_ATTR,
    ExifMarkers,
)
from exif._datatypes import (
    ExifType,
    Ifd,
    IfdLe,
    IfdTag,
    IfdTagLe,
    TiffByteOrder,
    TiffHeader,
)
from exif.ifd_tag import (
    Ascii,
    BaseIfdTag,
    Byte,
    ExifVersion,
    Long,
    Rational,
    Short,
    Slong,
    Srational,
    Sshort,
    UserComment,
    WindowsXp,
)
from exif.ifd_tag._rational import RationalDtype
from exif.ifd_tag._srational import SrationalDtype
from exif.ifd_tag._user_comment import USER_COMMENT_CHARACTER_CODE_LEN_BYTES

# FUTURE: There's quite a few spots where a new Plum Buffer is created for the APP1 body bytes. Consider cleaning this
# up to share the same buffer reference. Also, fix the false positive no-member Pylint errors.
# pylint: disable=no-member


class App1MetaData:

    """APP1 metadata interface class for EXIF tags."""

    def _add_empty_ifd(self, ifd):
        if not ifd == "gps":
            raise RuntimeError("only can add GPS IFD to image, not {0}".format(ifd))

        if 1 not in self.ifd_pointers:
            raise RuntimeError("can't yet add to images without a subsequent IFD 1")

        if self.endianness == TiffByteOrder.BIG:
            ifd_cls = Ifd
        else:
            ifd_cls = IfdLe

        new_app1_bytes = self.body_bytes[: self.ifd_pointers[1]]
        bytes_after_new_ifd = self.body_bytes[self.ifd_pointers[1] :]

        # Inert empty IFD.
        empty_ifd = Ifd(tags=[], next=0)
        new_app1_bytes += empty_ifd.ipack()

        # Touch up pointer to IFD 1 (which we already know exists).
        new_app1_bytes_buffer = Buffer(new_app1_bytes)
        new_app1_bytes_buffer.offset = self.ifd_pointers[0]

        ifd_zero = new_app1_bytes_buffer.unpack(ifd_cls)
        ifd_zero.next += empty_ifd.nbytes
        new_app1_bytes[
            self.ifd_pointers[0] : self.ifd_pointers[0] + ifd_zero.nbytes
        ] = ifd_zero.ipack()

        # Touch up IFD 1 pointers!
        after_new_ifd_buffer = Buffer(bytes_after_new_ifd)
        after_new_ifd_buffer.offset = 0
        ifd1 = after_new_ifd_buffer.unpack(ifd_cls)

        for tag_index in range(ifd1.count):
            tag_t = ifd1.tags[tag_index]
            is_value_in_ifd_tag_itself = value_fits_in_ifd_tag(tag_t)
            if (
                tag_t.tag_id in [ATTRIBUTE_ID_MAP["jpeg_interchange_format"]]
                or not is_value_in_ifd_tag_itself
            ):
                tag_t.value_offset += empty_ifd.nbytes
            ifd1.tags[tag_index] = tag_t

        pack_into(datatype=ifd1, buffer=bytes_after_new_ifd, offset=0)

        # Parse new bytes containing the additional placeholder IFD.
        self.body_bytes = new_app1_bytes + bytes_after_new_ifd
        self._parse_ifd_segments()

        # Adjust the size of the APP1 header to reflect the new length.
        app1_len = uint16.view(
            self.header_bytes, offset=2
        )  # 2 bytes into the header, i.e., right after the marker
        app1_len += empty_ifd.nbytes

        # Add pointer tag to IFD 0.
        offset_of_new_ifd = self.ifd_pointers[
            1
        ]  # IFD 1 is pushed back to after the new IFD tag that takes its place
        self._add_tag("_gps_ifd_pointer", offset_of_new_ifd)

    def _add_tag(
        self, tag, value
    ):  # pylint: disable=too-many-locals, too-many-branches, too-many-statements
        # FUTURE: This method could likely use some future cleanup and abstraction.
        try:
            tag_type, ifd_number = ATTRIBUTE_TYPE_MAP[tag]
        except KeyError:
            raise AttributeError("cannot add attribute {0} to image".format(tag))

        if self.endianness == TiffByteOrder.BIG:
            ifd_cls = Ifd
            ifd_tag_cls = IfdTag
        else:
            ifd_cls = IfdLe
            ifd_tag_cls = IfdTagLe

        if ifd_number not in self.ifd_pointers:
            self._add_empty_ifd(ifd_number)

        # Make a list of all IFDs that will need to be re-packed with touched up pointers.
        subsequent_ifd_names = [
            ifd
            for ifd, offset in self.ifd_pointers.items()
            if offset > self.ifd_pointers[ifd_number]
        ]
        subsequent_ifd_offsets = sorted(
            [
                offset
                for offset in self.ifd_pointers.values()
                if offset > self.ifd_pointers[ifd_number]
            ]
        )

        # Determine the number of bytes that will be injected.
        added_bytes = ifd_tag_cls.nbytes
        pointer_value_bytes = 0
        value_count = 1

        if tag_type == ExifType.ASCII and len(value) >= 4:
            pointer_value_bytes = len(value) + 1  # add one for null termination

        if tag_type == ExifType.ASCII:
            value_count = len(value) + 1

        if tag_type == ExifType.RATIONAL:
            if isinstance(value, tuple):
                value_count = len(value)
            else:
                value_count = 1

            pointer_value_bytes = value_count * RationalDtype.nbytes

        if tag_type == ExifType.SRATIONAL:
            # Value count stays at 1 since EXIF specification does not define multi-valued SRATIONAL tags.
            pointer_value_bytes = value_count * SrationalDtype.nbytes

        if (
            tag == "user_comment"
        ):  # character code header followed by null-terminated ASCII string
            value_count = USER_COMMENT_CHARACTER_CODE_LEN_BYTES + len(value) + 1

            if len(value) >= 4:
                pointer_value_bytes = (
                    USER_COMMENT_CHARACTER_CODE_LEN_BYTES + len(value) + 1
                )

        added_bytes += pointer_value_bytes

        # Keep all bytes prior to the IFD where the new tag will be added.
        new_app1_bytes = self.body_bytes[: self.ifd_pointers[ifd_number]]

        # If IFD 1 occurs after that added tag, adjust the pointer to it from IFD 0.
        if ifd_number != 0:
            new_app1_buffer = Buffer(new_app1_bytes)
            new_app1_buffer.offset = self.ifd_pointers[0]
            ifd_zero = new_app1_buffer.unpack(ifd_cls)

            if ifd_zero.next and 1 in subsequent_ifd_names:
                ifd_zero.next += added_bytes

            # Also adjust the pointers to the GPS and EXIF IFDs if they occur after the added tag.
            for tag_index in range(ifd_zero.count):
                tag_t = ifd_zero.tags[tag_index]

                is_ifd_pointer_to_adjust = (
                    tag_t.tag_id == ATTRIBUTE_ID_MAP["_gps_ifd_pointer"]
                    and "gps" in subsequent_ifd_names
                )
                is_ifd_pointer_to_adjust |= (
                    tag_t.tag_id == ATTRIBUTE_ID_MAP["_exif_ifd_pointer"]
                    and "exif" in subsequent_ifd_names
                )

                if is_ifd_pointer_to_adjust:
                    tag_t.value_offset += added_bytes

                ifd_zero.tags[tag_index] = tag_t

            pack_into(
                datatype=ifd_zero, buffer=new_app1_bytes, offset=self.ifd_pointers[0]
            )

        # Unpack the original bytes of the IFD to which the new tag will be added to.
        target_ifd_offset = self.ifd_pointers[ifd_number]
        body_bytes_buffer = Buffer(self.body_bytes)
        body_bytes_buffer.offset = target_ifd_offset
        target_ifd = body_bytes_buffer.unpack(ifd_cls)

        if subsequent_ifd_offsets:
            orig_ifd_values = self.body_bytes[
                target_ifd_offset + target_ifd.nbytes : subsequent_ifd_offsets[0]
            ]
        else:
            orig_ifd_values = self.body_bytes[target_ifd_offset + target_ifd.nbytes :]

        # Determine if a pointer to a value is necessary, and if so, find it.
        if (
            (tag_type == ExifType.ASCII and len(value) >= 4)
            or tag_type in [ExifType.RATIONAL, ExifType.SRATIONAL]
            or (tag == "user_comment" and len(value) >= 4)
        ):
            if subsequent_ifd_offsets:
                value_pointer = subsequent_ifd_offsets[0] + ifd_tag_cls.nbytes
            else:
                # Can put at end since if EXIF or GPS is the last IFD, there must not be a thumbnail and IFD 1.
                value_pointer = len(self.body_bytes) + ifd_tag_cls.nbytes
        elif tag == "_gps_ifd_pointer":
            # Must set pointer values now or else they'll incorrectly point to 0x00 when parsing.
            value_pointer = value
        else:
            value_pointer = 0

        # Iterate over the IFD's tags and increase any value offset pointers by the size of an IFD tag.
        for tag_index in range(target_ifd.count):
            tag_t = target_ifd.tags[tag_index]

            is_ifd_pointer_to_adjust = (
                tag_t.tag_id == ATTRIBUTE_ID_MAP["_gps_ifd_pointer"]
                and "gps" in subsequent_ifd_names
            )
            is_ifd_pointer_to_adjust |= (
                tag_t.tag_id == ATTRIBUTE_ID_MAP["_exif_ifd_pointer"]
                and "exif" in subsequent_ifd_names
            )

            is_value_in_ifd_tag_itself = value_fits_in_ifd_tag(tag_t)
            if is_ifd_pointer_to_adjust:
                tag_t.value_offset += added_bytes
            elif not is_value_in_ifd_tag_itself:
                tag_t.value_offset += ifd_tag_cls.nbytes

            target_ifd.tags[tag_index] = tag_t

        # Add the new tag to the IFD.
        target_ifd.count += 1
        target_ifd.tags.append(
            ifd_tag_cls(
                tag_id=ATTRIBUTE_ID_MAP[tag],
                type=tag_type,
                value_count=value_count,
                value_offset=value_pointer,
            )
        )

        # If necessary, touch up the pointer to the next IFD.
        if target_ifd.next:
            target_ifd.next += added_bytes

        # Pack new IFD bytes into the new body bytes (along with the pre-existing values that follow).
        new_app1_bytes[
            target_ifd_offset : target_ifd_offset + target_ifd.nbytes
        ] = target_ifd.ipack()
        new_app1_bytes += orig_ifd_values
        new_app1_bytes += b"\x00" * pointer_value_bytes

        # Touch up pointers in any subsequent IFDs.
        # FUTURE: This could likely be better abstracted!
        while subsequent_ifd_offsets:
            current_ifd_offset = subsequent_ifd_offsets.pop(0)

            body_bytes_buffer = Buffer(self.body_bytes)
            body_bytes_buffer.offset = current_ifd_offset
            target_ifd = body_bytes_buffer.unpack(ifd_cls)

            if subsequent_ifd_offsets:
                orig_ifd_values = self.body_bytes[
                    current_ifd_offset + target_ifd.nbytes : subsequent_ifd_offsets[0]
                ]
            else:
                orig_ifd_values = self.body_bytes[
                    current_ifd_offset + target_ifd.nbytes :
                ]

            for tag_index in range(target_ifd.count):
                tag_t = target_ifd.tags[tag_index]
                is_value_in_ifd_tag_itself = value_fits_in_ifd_tag(tag_t)
                if (
                    tag_t.tag_id in [ATTRIBUTE_ID_MAP["jpeg_interchange_format"]]
                    or not is_value_in_ifd_tag_itself
                ):
                    tag_t.value_offset += added_bytes

                target_ifd.tags[tag_index] = tag_t

            if target_ifd.next:
                target_ifd.next += added_bytes

            new_app1_bytes += target_ifd.ipack()
            new_app1_bytes += orig_ifd_values

        # Finally, adjust the size of the APP1 header to reflect the new length.
        app1_len = uint16.view(
            self.header_bytes, offset=2
        )  # 2 bytes into the header, i.e., right after the marker
        app1_len += added_bytes

        # Reload to pick up on new bytes arrangement and then modify the currently-zero value.
        self.body_bytes = new_app1_bytes
        self._parse_ifd_segments()
        self.ifd_tags[ATTRIBUTE_ID_MAP[tag]].modify(value)

        # If the tag is a user comment, update its character code header to reflect ASCII encoding.
        if tag == "user_comment":
            self.ifd_tags[ATTRIBUTE_ID_MAP[tag]].set_character_code_to_ascii()

    def _delete_ifd_tag(self, attribute_id):
        # Overwrite pointer data with null bytes (if applicable, depending on datatype).
        self.ifd_tags[attribute_id].wipe()

        # Unpack the original IFD section.
        corresponding_ifd_offset = self.ifd_pointers[self.tag_parent_ifd[attribute_id]]
        if self.endianness == TiffByteOrder.BIG:
            ifd_cls = Ifd
        else:
            ifd_cls = IfdLe

        body_bytes_buffer = Buffer(self.body_bytes)
        body_bytes_buffer.offset = corresponding_ifd_offset
        orig_ifd = body_bytes_buffer.unpack(ifd_cls)

        # Construct a new IFD section datatype containing all tags but the deletion target.
        preserved_tags = [tag for tag in orig_ifd.tags if tag.tag_id != attribute_id]
        new_ifd = ifd_cls(tags=preserved_tags, next=orig_ifd.next)

        # Pack in new IFD bytes with null bytes (i.e., an empty IFD tag) appended to preserve pointers.
        # Note: The pack_into method overrides the pre-existing bytes.
        pack_into(
            datatype=new_ifd, buffer=self.body_bytes, offset=corresponding_ifd_offset
        )
        pack_into(
            datatype=IfdTag(tag_id=0, type=0, value_count=0, value_offset=0),
            buffer=self.body_bytes,
            offset=corresponding_ifd_offset + new_ifd.nbytes,
        )

        # Remove tag from parser tag dictionary.
        del self.ifd_tags[attribute_id]
        del self.tag_parent_ifd[attribute_id]

        # Regenerate information about existing tags.
        self._parse_ifd_segments()

    def _extract_thumbnail(self):
        if 1 in self.ifd_pointers:  # IFD segment 1 contains thumbnail (if present)
            hex_after_ifd1 = self.body_bytes[self.ifd_pointers[1] :]
            try:
                start_index = hex_after_ifd1.index(ExifMarkers.SOI)
                end_index = hex_after_ifd1.index(ExifMarkers.EOI) + len(ExifMarkers.EOI)
            except ValueError:  # pragma: no cover
                pass  # no thumbnail
            else:
                self.thumbnail_bytes = hex_after_ifd1[start_index:end_index]

    def get_segment_bytes(self) -> bytes:
        """Get equivalent APP1 segment bytes."""
        return bytes(self.header_bytes) + bytes(self.body_bytes)

    def get_tag_list(self, include_unknown: bool = True) -> List[str]:
        """Get a list of EXIF tag attributes present in the image object."""
        if include_unknown:
            tag_list = [
                ATTRIBUTE_NAME_MAP.get(key, "<unknown EXIF tag {0}>".format(key))
                for key in self.ifd_tags
            ]

        else:
            tag_list = []

            for key in self.ifd_tags:
                try:
                    tag_list.append(ATTRIBUTE_NAME_MAP[key])
                except KeyError:
                    pass

        return tag_list

    def _iter_ifd_tags(self, ifd_key):
        ifd_offset = self.ifd_pointers[ifd_key]

        if self.endianness == TiffByteOrder.BIG:
            ifd_cls = Ifd
        else:
            ifd_cls = IfdLe

        try:
            body_bytes_buffer = Buffer(self.body_bytes)
            body_bytes_buffer.offset = ifd_offset
            ifd_t = body_bytes_buffer.unpack(ifd_cls)
        except UnpackError:
            warnings.warn(f"skipping bad IFD {ifd_key}", RuntimeWarning)
            next_ifd_offset = 0
        else:
            for tag_index in range(ifd_t.count):
                tag_offset = (
                    ifd_offset + 2 + tag_index * IfdTag.nbytes
                )  # count is 2 bytes
                tag_t = ifd_t.tags[tag_index]
                tag_py_ins = self._tag_factory(tag_t, tag_offset)

                if (
                    ifd_key != 1 or tag_t.tag_id not in self.ifd_tags
                ):  # don't let thumbnail tags override base image tags
                    self.ifd_tags[tag_t.tag_id] = tag_py_ins
                    self.tag_parent_ifd[tag_t.tag_id] = ifd_key

                if tag_t.tag_id == ATTRIBUTE_ID_MAP["_exif_ifd_pointer"]:
                    self.ifd_pointers["exif"] = tag_t.value_offset

                if tag_t.tag_id == ATTRIBUTE_ID_MAP["_gps_ifd_pointer"]:
                    self.ifd_pointers["gps"] = tag_t.value_offset

            next_ifd_offset = ifd_t.next

        return next_ifd_offset

    def _parse_ifd_segments(self):
        body_bytes_buffer = Buffer(self.body_bytes)
        tiff_header = body_bytes_buffer.unpack(TiffHeader)
        self.endianness = tiff_header.byte_order

        current_ifd = 0
        current_ifd_offset = tiff_header.ifd_offset

        while current_ifd_offset:
            self.ifd_pointers[current_ifd] = current_ifd_offset
            current_ifd_offset = self._iter_ifd_tags(current_ifd)
            current_ifd += 1

        if "exif" in self.ifd_pointers:
            self._iter_ifd_tags("exif")

        if "gps" in self.ifd_pointers:
            self._iter_ifd_tags("gps")

    def _tag_factory(self, tag_t, offset):  # pylint: disable=too-many-branches
        if (
            ATTRIBUTE_ID_MAP["xp_title"]
            <= tag_t.tag_id
            <= ATTRIBUTE_ID_MAP["xp_subject"]
        ):  # legacy Windows XP tags
            cls = WindowsXp
        elif (
            ATTRIBUTE_ID_MAP["exif_version"] == tag_t.tag_id
        ):  # custom ASCII encoding without termination character
            cls = ExifVersion
        elif ATTRIBUTE_ID_MAP["user_comment"] == tag_t.tag_id:
            cls = UserComment
        elif tag_t.type == ExifType.BYTE:
            cls = Byte
        elif tag_t.type == ExifType.ASCII:
            cls = Ascii
        elif tag_t.type == ExifType.SHORT:
            cls = Short
        elif tag_t.type == ExifType.LONG:
            cls = Long
        elif tag_t.type == ExifType.RATIONAL:
            cls = Rational
        elif tag_t.type == ExifType.SLONG:
            cls = Slong
        elif tag_t.type == ExifType.SRATIONAL:
            cls = Srational
        elif tag_t.type == ExifType.SSHORT:
            cls = Sshort
        else:
            cls = BaseIfdTag

        return cls(offset, self)

    def __init__(self, segment_bytes):
        self.header_bytes = bytearray(segment_bytes[:0xA])
        self.body_bytes = bytearray(segment_bytes[0xA:])

        self.endianness = None
        self.ifd_pointers = {}
        self.ifd_tags = {}
        self.tag_parent_ifd = {}
        self.thumbnail_bytes = None

        self._parse_ifd_segments()
        self._extract_thumbnail()

    def __delattr__(self, item):
        try:
            # Determine if attribute is an IFD tag accessor.
            attribute_id = ATTRIBUTE_ID_MAP[item]
        except KeyError:  # pragma: no cover
            # Coverage and behavior tested by Image class.
            # Attribute is a class member. Delete natively.
            super(App1MetaData, self).__delattr__(item)
        else:
            # Attribute is not a class member. Delete EXIF tag value.
            try:
                self.ifd_tags[attribute_id]
            except KeyError:
                raise AttributeError(ERROR_IMG_NO_ATTR.format(item))

            self._delete_ifd_tag(attribute_id)

    def __getattr__(self, item):
        """If attribute is not a class member, get the value of the EXIF tag of the same name."""
        try:
            attribute_id = ATTRIBUTE_ID_MAP[item.lower()]
        except KeyError:
            raise AttributeError("unknown image attribute {0}".format(item))

        try:
            ifd_tag = self.ifd_tags[attribute_id]
        except KeyError:
            raise AttributeError(ERROR_IMG_NO_ATTR.format(item))

        return ifd_tag.read()

    def __setattr__(self, key, value):
        try:
            # Determine if attribute is an IFD tag accessor.
            attribute_id = ATTRIBUTE_ID_MAP[key]
        except KeyError:
            # Attribute is a class member. Set natively.
            super(App1MetaData, self).__setattr__(key, value)
        else:
            try:
                ifd_tag = self.ifd_tags[attribute_id]
            except KeyError:
                # Tag is not in image already.
                self._add_tag(key, value)
            else:
                try:
                    ifd_tag.modify(value)
                except ValueError as exc:  # e.g., if doesn't fit into tag, try deleting and re-adding
                    try:
                        attr_exif_type = ATTRIBUTE_TYPE_MAP[key][0]
                    except KeyError:
                        raise exc

                    if attr_exif_type == ExifType.ASCII:
                        self._delete_ifd_tag(attribute_id)
                        self._add_tag(key, value)
                    else:
                        raise exc



"""IFD tag structure parser classes."""

from exif.ifd_tag._ascii import Ascii
from exif.ifd_tag._base import Base as BaseIfdTag
from exif.ifd_tag._byte import Byte
from exif.ifd_tag._exifversion import ExifVersion
from exif.ifd_tag._long import Long
from exif.ifd_tag._rational import Rational
from exif.ifd_tag._short import Short
from exif.ifd_tag._slong import Slong
from exif.ifd_tag._sshort import Sshort
from exif.ifd_tag._srational import Srational
from exif.ifd_tag._user_comment import UserComment
from exif.ifd_tag._windows_xp import WindowsXp



"""Package EXIF-specific datatypes."""

from enum import IntEnum, IntFlag

from plum.array import ArrayX
from plum.bigendian import uint16, uint32
from plum.bitfields import BitFields, bitfield
from plum.enum import EnumX
from plum.littleendian import uint16 as uint16_l, uint32 as uint32_l
from plum.structure import (
    member,
    sized_member,
    Structure,
)


class TiffByteOrder(IntEnum):

    """TIFF Header Byte Order Indicator"""

    LITTLE = 0x4949
    BIG = 0x4D4D


tiff_byte_order = EnumX(
    "tiff_byte_order",
    TiffByteOrder,
    nbytes=2,
    byteorder="big",
    signed=False,
    strict=True,
)
BYTE_ORDER_MAP = {TiffByteOrder.LITTLE: uint32_l, TiffByteOrder.BIG: uint32}


class TiffHeader(Structure):

    """TIFF Header"""

    byte_order: int = member(fmt=tiff_byte_order)
    reserved: int = member(fmt=uint16)
    ifd_offset: int = member(fmt=BYTE_ORDER_MAP.__getitem__, fmt_arg=byte_order)  # type: ignore


class ExifType(IntEnum):

    """EXIF Tag Types"""

    EMPTY = 0
    BYTE = 1
    ASCII = 2
    SHORT = 3
    LONG = 4
    RATIONAL = 5
    UNDEFINED = 7
    SSHORT = 8
    SLONG = 9
    SRATIONAL = 10


exif_type = EnumX("exif_type", ExifType, nbytes=2, byteorder="big", signed=False)
exif_type_le = EnumX("exif_type", ExifType, nbytes=2, byteorder="little", signed=False)


class IfdTag(Structure):

    """IFD Tag"""

    tag_id: int = member(fmt=uint16)
    type: int = member(fmt=exif_type)
    value_count: int = member(fmt=uint32)
    value_offset: int = member(fmt=uint32)


class IfdTagLe(Structure):

    """IFD Tag (Little Endian)"""

    tag_id: int = member(fmt=uint16_l)
    type: int = member(fmt=exif_type_le)
    value_count: int = member(fmt=uint32_l)
    value_offset: int = member(fmt=uint32_l)


ifd_tag_array = ArrayX("ifd_tag", fmt=IfdTag)
ifd_tag_array_le = ArrayX("ifd_tag_le", fmt=IfdTagLe)


class Ifd(Structure):

    """IFD Segment"""

    count: int = member(fmt=uint16, compute=True)  # type: ignore
    tags: list = sized_member(size=count, fmt=ifd_tag_array, ratio=IfdTag.nbytes)  # type: ignore
    next: int = member(fmt=uint32)


class IfdLe(Structure):

    """IFD Segment (Little Endian)"""

    count: int = member(fmt=uint16_l, compute=True)  # type: ignore
    tags: list = sized_member(size=count, fmt=ifd_tag_array_le, ratio=IfdTagLe.nbytes)  # type: ignore
    next: int = member(fmt=uint32_l)


class FlashReturn(IntFlag):

    """Flash status of returned light."""

    NO_STROBE_RETURN_DETECTION_FUNCTION = 0
    RESERVED = 1
    STROBE_RETURN_LIGHT_DETECTED = 2
    STROBE_RETURN_LIGHT_NOT_DETECTED = 3


class FlashMode(IntFlag):

    """Flash mode of the camera."""

    UNKNOWN = 0
    COMPULSORY_FLASH_FIRING = 1
    COMPULSORY_FLASH_SUPPRESSION = 2
    AUTO_MODE = 3


class Flash(BitFields, nbytes=1):  # type: ignore

    """Status of the camera's flash when the image was taken. (Reported by the ``flash`` tag.)"""

    flash_fired: bool = bitfield(typ=bool, size=1)
    flash_return: FlashReturn = bitfield(typ=FlashReturn, size=2)
    flash_mode: FlashMode = bitfield(typ=FlashMode, size=2)
    flash_function_not_present: bool = bitfield(typ=bool, size=1)
    red_eye_reduction_supported: bool = bitfield(typ=bool, size=1)
    reserved: int = bitfield(typ=int, size=1)



from setuptools import setup, find_packages

with open("README.rst", "r") as fh:
    long_description = fh.read()

with open("version", "r") as fh:
    version = fh.read()

setup(
    python_requires=">=3.7",
    install_requires=[
        "plum-py>=0.5.0",
    ],
    include_package_data=True,
    name="exif",
    version=version,
    author="Tyler N. Thieding",
    author_email="python@thieding.com",
    maintainer="Tyler N. Thieding",
    maintainer_email="python@thieding.com",
    url="https://gitlab.com/TNThieding/exif",
    description="Read and modify image EXIF metadata using Python.",
    long_description=long_description,
    download_url="https://gitlab.com/TNThieding/exif",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Multimedia :: Graphics :: Editors",
    ],
    package_dir={"": "src"},
    packages=find_packages("src"),
)
